#ifndef __IOKIT_IOCFBUNDLE_H
#define __IOKIT_IOCFBUNDLE_H

#include <CoreFoundation/CoreFoundation.h>

#define kIOBundleInfoDictionaryVersionKey	CFSTR("CFBundleInfoDictionaryVersion")
#define kIOBundleExecutableKey			CFSTR("CFBundleExecutable")
#define kIOBundleIdentifierKey			CFSTR("CFBundleIdentifier")
#define kIOBundleVersionKey			CFSTR("CFBundleVersion")
#define kIOBundleDevelopmentRegionKey		CFSTR("CFBundleDevelopmentRegion")
#define kIOBundleNameKey			CFSTR("CFBundleName")

#endif /* __IOKIT_IOCFBUNDLE_H */
